from django.shortcuts import get_object_or_404, redirect, render, resolve_url

from flags.models import FlagState
from flags.sources import get_flags
from flags.state import flag_enabled
from flags.templatetags.flags_debug import bool_enabled
from wagtailflags.forms import FlagStateForm, NewFlagForm


def index(request):
    context = {
        'flags': sorted(get_flags().values(), key=lambda x: x.name),
    }
    return render(request, 'wagtailflags/index.html', context)


def create_flag(request):
    """ Create a new flag.
    This will add a FlagState object with a custom name and a boolean: False
    condition """
    if request.method == 'POST':
        form = NewFlagForm(
            request.POST,
            initial={
                'condition': 'path matches',
                'value': '/foo',
                'required': False,
            }
        )
        if form.is_valid():
            form.save()
            if flag_enabled('WAGTAILFLAGS_ADMIN_BIG_LIST'):
                return redirect(
                    '{flags_list}#{flag_name}'.format(
                        flags_list=resolve_url('wagtailflags:list'),
                        flag_name=form.instance.name
                    )
                )
            else:
                return redirect(
                    'wagtailflags:flag_index', name=form.instance.name
                )
    else:
        form = NewFlagForm()

    context = dict(form=form)
    return render(request, 'wagtailflags/flags/create_flag.html', context)


def flag_index(request, name):
    flag = get_flags().get(name)

    # If there's a database boolean condition, fetch it and treat it as a
    # on/off switch
    if 'enable' in request.GET or 'disable' in request.GET:
        db_boolean_condition = next(
            (
                c for c in flag.conditions
                if c.condition == 'boolean'
                and getattr(c, 'obj', None) is not None
            ),
            None
        )

        if db_boolean_condition is None:
            boolean_condition_obj = FlagState.objects.create(
                name=name,
                condition='boolean',
                value='True'
            )
        else:
            boolean_condition_obj = db_boolean_condition.obj

        if 'enable' in request.GET and not bool_enabled(flag):
            boolean_condition_obj.value = True
        elif 'disable' in request.GET and bool_enabled(flag):
            boolean_condition_obj.value = False

        boolean_condition_obj.save()

        if flag_enabled('WAGTAILFLAGS_ADMIN_BIG_LIST'):
            return redirect(
                '{flags_list}#{flag_name}'.format(
                    flags_list=resolve_url('wagtailflags:list'),
                    flag_name=name
                )
            )
        else:
            return redirect('wagtailflags:flag_index', name=name)

    context = {
        'flag': flag,
    }
    return render(request, 'wagtailflags/flags/flag_index.html', context)


def edit_condition(request, name, condition_pk=None):
    flag = get_flags().get(name)
    try:
        condition = FlagState.objects.get(pk=condition_pk)
    except FlagState.DoesNotExist:
        condition = None

    if request.method == 'POST':
        form = FlagStateForm(
            request.POST, initial={'name': name}, instance=condition
        )
        if form.is_valid():
            form.save()

            if flag_enabled('WAGTAILFLAGS_ADMIN_BIG_LIST'):
                return redirect(
                    '{flags_list}#{flag_name}'.format(
                        flags_list=resolve_url('wagtailflags:list'),
                        flag_name=form.instance.name
                    )
                )
            else:
                return redirect(
                    'wagtailflags:flag_index', name=form.instance.name
                )

    else:
        form = FlagStateForm(initial={'name': name}, instance=condition)

    context = {
        'flag': flag,
        'form': form,
        'condition_str': str(condition),
        'condition_pk': condition_pk,
    }
    return render(request, 'wagtailflags/flags/edit_condition.html', context)


def delete_condition(request, name, condition_pk):
    flag = get_flags().get(name)
    condition = get_object_or_404(FlagState, pk=condition_pk)

    if request.method == 'POST':
        condition.delete()

        if flag_enabled('WAGTAILFLAGS_ADMIN_BIG_LIST'):
            return redirect(
                '{flags_list}#{flag_name}'.format(
                    flags_list=resolve_url('wagtailflags:list'),
                    flag_name=name
                )
            )
        else:
            return redirect('wagtailflags:flag_index', name=name)

    context = {
        'flag': flag,
        'condition_str': str(condition),
        'condition_pk': condition.pk,
    }
    return render(request, 'wagtailflags/flags/delete_condition.html', context)



from django.test import TestCase, override_settings

from wagtail.tests.utils import WagtailTestUtils

from flags.models import FlagState


class TestWagtailFlagsViews(TestCase, WagtailTestUtils):

    def setUp(self):
        self.login()

        self.dbonly_flag = FlagState.objects.create(
            name='DBONLY_FLAG',
            condition='boolean',
            value='True',
            required=False
        )

    def test_flags_index(self):
        response = self.client.get('/admin/flags/')
        self.assertEqual(response.status_code, 200)

        self.assertContains(response, 'FLAG_ENABLED')
        self.assertContains(response, 'FLAG_DISABLED')
        self.assertContains(response, 'DBONLY_FLAG')
        self.assertContains(response, '<b>enabled</b> when')
        self.assertContains(response, '<b>enabled</b> for')

    def test_flag_create(self):
        response = self.client.get('/admin/flags/create/')
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            '/admin/flags/create/', {'name': 'NEW_FLAG'}
        )
        self.assertRedirects(response, '/admin/flags/NEW_FLAG/')
        self.assertEqual(FlagState.objects.count(), 2)
        new_flag_condition = FlagState.objects.get(name='NEW_FLAG')
        self.assertEqual(new_flag_condition.condition, 'boolean')
        self.assertEqual(new_flag_condition.value, 'False')
        self.assertEqual(new_flag_condition.required, False)

    @override_settings(
        FLAGS={'WAGTAILFLAGS_ADMIN_BIG_LIST': [('boolean', True)]}
    )
    def test_flag_create_big_list(self):
        response = self.client.post(
            '/admin/flags/create/', {'name': 'NEW_FLAG'}
        )
        self.assertRedirects(response, '/admin/flags/#NEW_FLAG')

    def test_flag_create_existing(self):
        response = self.client.get('/admin/flags/create/')
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            '/admin/flags/create/', {'name': 'DBONLY_FLAG'}
        )
        self.assertContains(response, 'Flag named DBONLY_FLAG already exists')

    def test_flag_index(self):
        response = self.client.get('/admin/flags/FLAG_DISABLED/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'FLAG_DISABLED')
        self.assertContains(response, 'path matches')
        self.assertContains(response, 'Enable FLAG_DISABLED')

    def test_flag_index_enabled(self):
        response = self.client.get('/admin/flags/FLAG_ENABLED/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Disable FLAG_ENABLED')

    def test_flag_index_disabled(self):
        response = self.client.get('/admin/flags/FLAG_DISABLED/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Enable FLAG_DISABLED')

    def test_flag_index_enabled_required_true(self):
        response = self.client.get('/admin/flags/FLAG_ENABLED/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Disable FLAG_ENABLED')

    def test_flag_index_disabled_required_true(self):
        response = self.client.get('/admin/flags/FLAG_ENABLED/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Disable FLAG_ENABLED')

    def test_enable_flag(self):
        condition_query = FlagState.objects.filter(name='FLAG_DISABLED')
        self.assertEqual(len(condition_query.all()), 0)

        response = self.client.get(
            '/admin/flags/FLAG_DISABLED/', {'enable': ''}
        )
        self.assertRedirects(response, '/admin/flags/FLAG_DISABLED/')
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, 'boolean')
        self.assertEqual(condition_query.first().value, 'True')

    @override_settings(
        FLAGS={
            'WAGTAILFLAGS_ADMIN_BIG_LIST': [('boolean', True)],
            'FLAG_DISABLED': [],
        }
    )
    def test_enable_flag_big_list(self):
        response = self.client.get(
            '/admin/flags/FLAG_DISABLED/', {'enable': ''}
        )
        self.assertRedirects(response, '/admin/flags/#FLAG_DISABLED')

    def test_enable_flag_with_required_true(self):
        condition_query = FlagState.objects.filter(name='FLAG_DISABLED')
        self.assertEqual(len(condition_query.all()), 0)

        self.client.get('/admin/flags/FLAG_DISABLED/', {'enable': ''})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, 'boolean')
        self.assertEqual(condition_query.first().value, 'True')

    def test_disable_flag(self):
        condition_query = FlagState.objects.filter(name='DBONLY_FLAG')
        self.client.get('/admin/flags/DBONLY_FLAG/', {'disable': ''})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, 'boolean')
        self.assertEqual(condition_query.first().value, 'False')

    def test_disable_flag_with_required_true(self):
        condition_query = FlagState.objects.filter(name='DBONLY_FLAG')
        self.client.get('/admin/flags/DBONLY_FLAG/', {'disable': ''})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, 'boolean')
        self.assertEqual(condition_query.first().value, 'False')

    def test_create_flag_condition(self):
        response = self.client.get('/admin/flags/DBONLY_FLAG/create/')
        self.assertEqual(response.status_code, 200)

        params = {
            'condition': 'path matches',
            'value': '/db_path',
        }
        response = self.client.post('/admin/flags/DBONLY_FLAG/create/', params)
        self.assertRedirects(response, '/admin/flags/DBONLY_FLAG/')
        self.assertEqual(len(FlagState.objects.all()), 2)

    @override_settings(
        FLAGS={'WAGTAILFLAGS_ADMIN_BIG_LIST': [('boolean', True)]}
    )
    def test_create_flag_condition_big_list(self):
        params = {
            'condition': 'path matches',
            'value': '/db_path',
        }
        response = self.client.post('/admin/flags/DBONLY_FLAG/create/', params)
        self.assertRedirects(response, '/admin/flags/#DBONLY_FLAG')

    def test_edit_flag_condition(self):
        condition_obj = FlagState.objects.create(
            name='DBONLY_FLAG',
            condition='user',
            value='liberty'
        )
        response = self.client.get(
            '/admin/flags/DBONLY_FLAG/{}/'.format(condition_obj.pk)
        )
        self.assertEqual(response.status_code, 200)

        params = {
            'condition': 'user',
            'value': 'justice',
        }

        response = self.client.post(
            '/admin/flags/DBONLY_FLAG/{}/'.format(condition_obj.pk),
            params
        )
        self.assertRedirects(response, '/admin/flags/DBONLY_FLAG/')
        self.assertEqual(
            FlagState.objects.get(pk=condition_obj.pk).value,
            'justice'
        )

    @override_settings(
        FLAGS={'WAGTAILFLAGS_ADMIN_BIG_LIST': [('boolean', True)]}
    )
    def test_edit_flag_condition_big_list(self):
        condition_obj = FlagState.objects.create(
            name='DBONLY_FLAG',
            condition='user',
            value='liberty'
        )
        params = {
            'condition': 'user',
            'value': 'justice',
        }
        response = self.client.post(
            '/admin/flags/DBONLY_FLAG/{}/'.format(condition_obj.pk),
            params
        )
        self.assertRedirects(response, '/admin/flags/#DBONLY_FLAG')

    def test_delete_flag_condition(self):
        condition_obj = FlagState.objects.create(
            name='DBONLY_FLAG',
            condition='user',
            value='liberty'
        )
        self.assertEqual(len(FlagState.objects.all()), 2)
        response = self.client.get(
            '/admin/flags/DBONLY_FLAG/{}/delete/'.format(condition_obj.pk)
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            '/admin/flags/DBONLY_FLAG/{}/delete/'.format(condition_obj.pk)
        )
        self.assertRedirects(response, '/admin/flags/DBONLY_FLAG/')
        self.assertEqual(len(FlagState.objects.all()), 1)

    @override_settings(
        FLAGS={'WAGTAILFLAGS_ADMIN_BIG_LIST': [('boolean', True)]}
    )
    def test_delete_flag_condition_big_list(self):
        condition_obj = FlagState.objects.create(
            name='DBONLY_FLAG',
            condition='user',
            value='liberty'
        )
        self.assertEqual(len(FlagState.objects.all()), 2)
        response = self.client.post(
            '/admin/flags/DBONLY_FLAG/{}/delete/'.format(condition_obj.pk)
        )
        self.assertRedirects(response, '/admin/flags/#DBONLY_FLAG')



from django import forms

from flags.conditions import get_conditions
from flags.models import FlagState
from flags.sources import get_flags


class NewFlagForm(forms.ModelForm):
    name = forms.CharField(
        label='Name',
        required=True
    )

    def clean_name(self):
        name = self.cleaned_data['name']
        if name in get_flags():
            raise forms.ValidationError(
                "Flag named {} already exists".format(name)
            )
        return name

    def save(self, commit=True):
        obj = super(NewFlagForm, self).save(commit=False)
        obj.condition = 'boolean'
        obj.value = 'False'
        obj.required = False
        obj.save()
        return obj

    class Meta:
        model = FlagState
        fields = ('name', )


class FlagStateForm(forms.ModelForm):
    name = forms.CharField(
        label='Flag',
        required=True,
        disabled=True,
    )
    condition = forms.ChoiceField(
        label='Condition name',
        required=True
    )
    value = forms.CharField(
        label='Expected value',
        required=True
    )
    required = forms.BooleanField(
        label='Required',
        required=False,
        help_text=('All conditions marked "required" must be met to enable '
                   'the flag'),
    )

    def __init__(self, *args, **kwargs):
        super(FlagStateForm, self).__init__(*args, **kwargs)

        self.fields['condition'].choices = [
            (c, c) for c in sorted(get_conditions()) if c != 'boolean'
        ]

    class Meta:
        model = FlagState
        fields = ('name', 'condition', 'value', 'required')



from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

from flags.conditions import RequiredForCondition, register


@register('site')
def site_condition(site_str, request=None, **kwargs):
    """ Does the requests's Wagtail Site match the given site?
    site_str should be 'hostname:port', or 'hostname [default]'. """
    if request is None:
        raise RequiredForCondition("request is required for condition "
                                   "'site'")

    Site = apps.get_model('wagtailcore.Site')

    if '[default]' in site_str:
        # Wagtail Sites on the default port have [default] at the end of
        # their str() form.
        site_str = site_str.replace(' [default]', ':80')
    elif ':' not in site_str:
        # Add a default port if one isn't given
        site_str += ':80'

    hostname, port = site_str.split(':')
    try:
        conditional_site = Site.objects.get(hostname=hostname, port=port)
    except ObjectDoesNotExist:
        return False

    site = Site.find_for_request(request)

    return conditional_site == site



from django import template

from flags.templatetags.flags_debug import bool_enabled


register = template.Library()


@register.filter
def enablable(flag):
    return (
        not any(
            c.required for c in flag.conditions if c.condition == 'boolean'
        )
        and not bool_enabled(flag)
    )


@register.filter
def disablable(flag):
    return (
        not any(
            c.required for c in flag.conditions if c.condition == 'boolean'
        )
        and bool_enabled(flag)
    )



from django.test import TestCase, override_settings

from flags.sources import get_flags
from wagtailflags.templatetags.wagtailflags_admin import disablable, enablable


class TestWagtailFlagsAdminTemplateTags(TestCase):

    @override_settings(FLAGS={'MYFLAG': [('boolean', False)]})
    def test_enablable_disabled_not_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertTrue(enablable(flag))

    @override_settings(FLAGS={'MYFLAG': [('boolean', True)]})
    def test_enablable_enabled_not_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertFalse(enablable(flag))

    @override_settings(FLAGS={'MYFLAG': [('boolean', True, True)]})
    def test_enablable_disabled_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertFalse(enablable(flag))

    @override_settings(FLAGS={'MYFLAG': [('boolean', True)]})
    def test_disablable_disabled_not_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertTrue(disablable(flag))

    @override_settings(FLAGS={'MYFLAG': [('boolean', False)]})
    def test_disablable_enabled_not_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertFalse(disablable(flag))

    @override_settings(FLAGS={'MYFLAG': [('boolean', False, True)]})
    def test_disablable_enabled_required(self):
        flag = get_flags().get('MYFLAG')
        self.assertFalse(disablable(flag))



from django.apps import AppConfig


class WagtailFlagsAppConfig(AppConfig):
    name = 'wagtailflags'
    verbose_name = 'Wagtail Flags'

    def ready(self):
        # Import custom conditions to ensure that they get registered.
        import wagtailflags.conditions  # noqa



from django.http import HttpRequest
from django.test import TestCase

from flags.conditions import RequiredForCondition
from wagtailflags.conditions import site_condition


try:
    from wagtail.core.models import Site
except ImportError:  # pragma: no cover; fallback for Wagtail < 2.0
    from wagtail.wagtailcore.models import Site


class SiteConditionTestCase(TestCase):

    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.request = HttpRequest()
        self.request.site = self.site

    def test_site_valid_string(self):
        self.assertTrue(site_condition('localhost:80', request=self.request))

    def test_site_valid_string_no_port(self):
        self.assertTrue(site_condition('localhost', request=self.request))

    def test_site_valid_string_default_port(self):
        self.assertTrue(site_condition('localhost [default]',
                                       request=self.request))

    def test_site_valid_site(self):
        self.assertTrue(site_condition(str(self.site), request=self.request))

    def test_site_invalid_site(self):
        self.assertFalse(site_condition('non.existent.site',
                                        request=self.request))

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            site_condition('localhost:80')



from django.test import TestCase

from flags.conditions import get_conditions


class TestWagtailFlagsApps(TestCase):

    def test_condition_registration(self):
        self.assertIn('site', get_conditions())



import django
from django.conf.urls import include, url
from django.templatetags.static import static
from django.utils.html import format_html

from wagtailflags import views


try:  # pragma: no cover; >= 2.0
    from django.urls import reverse
except ImportError:  # pragma: no cover; fallback for Django < 2.0
    from django.core.urlresolvers import reverse

try:  # pragma: no cover; Wagtail >= 2.0
    from wagtail.admin.menu import MenuItem
    from wagtail.core import hooks
except ImportError:  # pragma: no cover; fallback for Wagtail < 2.0
    from wagtail.wagtailadmin.menu import MenuItem
    from wagtail.wagtailcore import hooks


@hooks.register('register_settings_menu_item')
def register_flags_menu():
    return MenuItem('Flags', reverse('wagtailflags:list'),
                    classnames='icon icon-tag', order=10000)


@hooks.register('register_admin_urls')
def register_flag_admin_urls():
    flagpatterns = [
        url(r'^$', views.index, name='list'),
        url(r'^create/$', views.create_flag, name='create_flag'),
        url(r'^(?P<name>[\w\-]+)/$', views.flag_index, name='flag_index'),
        url(
            r'^(?P<name>[\w\-]+)/create/$',
            views.edit_condition,
            name='create_condition'
        ),
        url(
            r'^(?P<name>[\w\-]+)/(?P<condition_pk>\d+)/$',
            views.edit_condition,
            name='edit_condition'
        ),
        url(
            r'^(?P<name>[\w\-]+)/(?P<condition_pk>\d+)/delete/$',
            views.delete_condition,
            name='delete_condition'
        ),
    ]

    if django.VERSION >= (1, 10):  # pragma: no cover
        urlpatterns = [
            url(r'^flags/',
                include((flagpatterns, 'wagtailflags'),
                        namespace='wagtailflags'))
        ]
    else:  # pragma: no cover; fallback for Django < 1.10
        urlpatterns = [
            url(r'^flags/',
                include((flagpatterns, 'wagtailflags', 'wagtailflags')))
        ]

    return urlpatterns


@hooks.register("insert_global_admin_css", order=100)
def global_admin_css():
    return format_html(
        '<link rel="stylesheet" href="{}">',
        static('wagtailflags/css/wagtailflags.css')
    )



from setuptools import find_packages, setup


long_description = open('README.md', 'r').read()

install_requires = [
    'wagtail>=1.13,<2.5',
    'django-flags>=4.2,<5.0'
]

testing_extras = [
    'coverage>=3.7.0',
]

setup(
    name='wagtail-flags',
    url='https://github.com/cfpb/wagtail-flags',
    author='CFPB',
    author_email='tech@cfpb.gov',
    description='Feature flags for Wagtail sites',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='CC0',
    version='4.1.0b2',
    include_package_data=True,
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        'testing': testing_extras,
    },
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Wagtail',
        'Framework :: Wagtail :: 1',
        'Framework :: Wagtail :: 2',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'License :: Public Domain',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



from django.conf.urls import include, url


try:  # pragma: no cover; Wagtail >= 2.0
    from wagtail.admin import urls as wagtailadmin_urls
except ImportError:  # pragma: no cover; fallback for Wagtail < 2.0
    from wagtail.wagtailadmin import urls as wagtailadmin_urls


urlpatterns = [
    url(r'^admin/', include(wagtailadmin_urls)),
]
